! WHIZARD 1.11 May 16 2001
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module whizard

  use kinds, only: default !NODEP!
  use monte_carlo_states
  use limits, only: DEFAULT_FILENAME, FILENAME_LEN
  use analyze
  use mpi90 !NODEP!
  use limits, only: WHIZARD_ROOT
  use vamp_interface, only: mc_integrate
  use vamp_interface, only: mc_generate
  use file_utils !NODEP!
  use vamp, only: vamp_copy_grids !NODEP!
  use stdhep_interface !NODEP!
  use jetset_interface !NODEP!
  use hepevt_common
  use events, only: write, RAW_FMT
  use histograms !NODEP!
  use limits, only: WEIGHT_HISTOGRAMS_NBIN
  use events, only: JETSET_FMT1, JETSET_FMT2, JETSET_FMT3, STDHEP_FMT

  implicit none
  private

  public :: whizard_set_filename
  public :: whizard_read_input
  public :: whizard_write_output
  public :: whizard_integrate
  public :: whizard_get_results
  public :: whizard_generate
  public :: whizard_events_begin
  public :: whizard_event
  public :: whizard_events_end
  public :: whizard_end

  public :: whizard_input

  type(monte_carlo_input) :: whizard_input
  character(len=FILENAME_LEN) :: whizard_filename = DEFAULT_FILENAME
  type(analysis_bundle) :: whizard_analysis
  logical :: whizard_input_exists = .false.
  logical :: mcs_exists = .false.
  logical :: event_files_open = .false.
  logical :: mpi_initialized = .false.
  integer :: proc_id
  integer :: uraw, uevt, uwgt

  save

contains

  subroutine whizard_set_filename (filename)
    character(len=*), intent(in) :: filename
    if (len(filename) > FILENAME_LEN) then
       call mpi90_rank (proc_id)
       if (proc_id == WHIZARD_ROOT) &
            & print *, "! *** Warning: filename too long; will be truncated."
    end if
    whizard_filename = trim(filename)
  end subroutine whizard_set_filename

  subroutine whizard_read_input
    if (whizard_input_exists .or. mcs_exists)  call whizard_end
    if (.not.mpi_initialized)  then
       call mpi90_init;  mpi_initialized = .true.
    end if
    call mpi90_rank (proc_id)
    call create (whizard_input)
    if (proc_id == WHIZARD_ROOT) then
       print *, "! WHIZARD 1.11 (May 16 2001)"
       call read (trim(whizard_filename), whizard_input)
    end if
    call mpi90_broadcast (whizard_input, WHIZARD_ROOT)
    whizard_input_exists = .true.
  end subroutine whizard_read_input

  subroutine whizard_write_output
    call mpi90_rank (proc_id)
    if (proc_id == WHIZARD_ROOT) then
       if (whizard_input_exists) then
          if (mcs_exists) then
             call write (trim(whizard_filename), mcs, whizard_input)
          else
             call create (mcs, whizard_input, trim(whizard_filename))
             call write (trim(whizard_filename), mcs, whizard_input)
             call destroy (mcs)
          end if
       else
          print *, "! Warning: Output file requested before input is known.  Ignored."
       end if
    end if
  end subroutine whizard_write_output

  subroutine whizard_integrate
    real(kind=default) :: time_per_event
    if (mcs_exists)  call whizard_end
    if (.not.whizard_input_exists)  call whizard_read_input
    call mpi90_rank (proc_id)
    call create (mcs, whizard_input, trim(whizard_filename))
    mcs_exists = .true.
    call mc_integrate (whizard_input, filename=trim(whizard_filename), &
         &             time_per_event = time_per_event)
    if (proc_id == WHIZARD_ROOT) then
       if (mcs%screen_diagnostics)  call write (6, mcs, whizard_input)
       if (mcs%screen_results)  call show_time_estimate
    end if
  contains
    subroutine show_time_estimate
      if (time_per_event > 0) then
         print *, "!"
         print "(1x,A,I9,A,I2.2,A,I2.2,A)", &
             & "! Time estimate for generating 10000 unweighted events:", &
             & int(10000*time_per_event)/3600, "h ", &
             & mod(int(10000*time_per_event), 3600)/60, "m ", &
             & mod(int(10000*time_per_event), 60), "s" 
      end if
    end subroutine show_time_estimate
  end subroutine whizard_integrate

  subroutine whizard_get_results (integral, error, chi2, efficiency)
    real(kind=default), intent(out) :: integral, error, chi2, efficiency
    if (mcs_exists) then
       integral = mcs%integral
       error = mcs%error
       chi2 = mcs%chi2
       efficiency = mcs%efficiency
    else
       call mpi90_rank (proc_id)
       if (proc_id == WHIZARD_ROOT) &
            & print *, "! Warning: Integration results requested but not known."
       integral = 0
       error = 0
       chi2 = 0
       efficiency = 0
    end if
  end subroutine whizard_get_results

  subroutine whizard_generate
    integer :: ievt, n_calls
    real(kind=default) :: lumi
    call mpi90_rank (proc_id)
    call get_n_calls (n_calls, lumi)
    if (n_calls > 0) then
       call whizard_events_begin (n_calls)
       if (mcs%unweighted) then
          if (proc_id == WHIZARD_ROOT) then
             print "(1x,A,1x,G11.4)", &
                  & "! Event sample corresponds to luminosity [fb-1] =", lumi
             if (mcs%read_events_raw) then
                print *, "! Reading", n_calls, "unweighted events ..."
             else
                print *, "! Generating", n_calls, "unweighted events ..."
             end if
          end if
          do ievt=1, n_calls
             call whizard_event
          end do
       else
          if (proc_id == WHIZARD_ROOT) then
             if (mcs%read_events_raw) then
                print *, "! Reading", n_calls, "weighted events ..."
             else
                print *, "! Generating", n_calls, "weighted events ..."
             end if
          end if
          call generate_events_weighted (n_calls)
       end if
       call whizard_events_end
    else
       if (proc_id == WHIZARD_ROOT)  print *, "! No event generation requested"
    end if
  contains
    subroutine get_n_calls (n_calls, lumi)
      integer, intent(out) :: n_calls
      real(kind=default), intent(out) :: lumi
      if (mcs%integral > 0) then
         if (mcs%unweighted) then
            lumi = max (mcs%luminosity, mcs%calls(2,5) / mcs%integral)
            n_calls = mcs%calls(1,5) * nint (mcs%integral * lumi)
         else
            if (mcs%luminosity > 0) then
               if (proc_id == WHIZARD_ROOT) &
                    & print *, "! Warning: Luminosity ignored for weighted events"
               lumi = 0
            end if
            n_calls = mcs%calls(1,5) * mcs%calls(2,5)
         end if
      else
         lumi = 0
         n_calls = 0
      end if
    end subroutine get_n_calls
  end subroutine whizard_generate

  subroutine whizard_events_begin (n_calls)
    integer, intent(in) :: n_calls
    if (event_files_open)  call whizard_events_end
    if (.not.mcs_exists)  call whizard_integrate
    call mpi90_rank (proc_id)
    if (mcs%integral <= 0) stop "Integral must be positive for simulation"
    call vamp_copy_grids (mcs%g, mcs%g_best)
    call create (whizard_analysis)
    if (proc_id == WHIZARD_ROOT) then
       call read (trim(whizard_filename), whizard_analysis, &
            &     trim(mcs%process_id))
    end if
!    call mpi90_broadcast (whizard_analysis, WHIZARD_ROOT)
    call open_raw_events_file (n_calls)
    call open_output_events_file (n_calls)
    if (mcs%fragment .and. mcs%fragmentation_method == JETSET) then
       if (.not.mcs%show_pythia_banner)  call jetset_suppress_banner
    else
       call jetset_suppress_banner
    end if
    call jetset_init (mcs%pygive)
    mcs%evt%count = 0
    event_files_open = .true.
  contains
    subroutine open_raw_events_file (n_calls)
      integer, intent(in) :: n_calls
      integer :: chsum1, chsum2
      if (mcs%read_events_raw) then
         if (proc_id == WHIZARD_ROOT) then
            print *, "! Looking for raw event file "// &
                 & trim(whizard_filename)//".evx ..."
            inquire(file=trim(whizard_filename)//".evx", &
                 &  exist=mcs%read_events_raw)
            if (.not.mcs%read_events_raw) &
                 & print *, "! ... not found.  Will be generated."
         end if
         call mpi90_broadcast (mcs%read_events_raw, WHIZARD_ROOT)
      end if
      if (mcs%unweighted) then
         chsum1 = mcs%checksum
      else
         chsum1 = mcs%checksum + 13579*n_calls
      end if
      if (mcs%read_events_raw) then
         if (proc_id == WHIZARD_ROOT) then
            uraw = free_unit()
            open (unit=uraw, file=trim(whizard_filename)//".evx", &
                 & status='old', action='read', form='unformatted')
            read(uraw) chsum2
         end if
      else if (mcs%write_events_raw) then
         if (proc_id == WHIZARD_ROOT) then
            uraw = free_unit()
            open(unit=uraw, file=trim(whizard_filename)//".evx", &
                 & status='replace', action='write', form='unformatted')
            chsum2 = chsum1
            write(uraw) chsum2
         end if
      end if
      call mpi90_broadcast (chsum1, WHIZARD_ROOT)
      call mpi90_broadcast (chsum2, WHIZARD_ROOT)
      if (.not.mcs%read_events_force .and. chsum1 /= chsum2) &
           & call checksum_error
    end subroutine open_raw_events_file
    subroutine open_output_events_file (n_calls)
      integer, intent(in) :: n_calls
      if (proc_id == WHIZARD_ROOT .and. mcs%write_events) then
         if (mcs%unweighted) then
            select case (mcs%write_events_format)
            case (JETSET_FMT1, JETSET_FMT2, JETSET_FMT3)
               print *, "! Writing events to standard output"
            case (STDHEP_FMT)
               print *, "! Writing events to file "// &
                    & trim(whizard_filename)//".evt"
               call stdhep_init (trim(whizard_filename)//".evt", &
                    &            "WHIZARD event sample", n_calls)
            case default
               print *, "! Writing events to file "// &
                    & trim(whizard_filename)//".evt"
               uevt = free_unit()
               open(unit=uevt, file=trim(whizard_filename)//".evt", &
                    & status='replace', action='write')
            end select
         else
            print *, "! Warning: can't write weighted events to file"// &
                 & trim(whizard_filename)//".evt"
         end if
      end if
    end subroutine open_output_events_file
  end subroutine whizard_events_begin

  subroutine whizard_event
    real(kind=default) :: excess
    integer :: iostat
    call mpi90_rank (proc_id)
    if (.not.event_files_open) &
         & stop "Missing call to whizard_events_begin"
    excess = 0
    if (mcs%read_events_raw) then
       if (proc_id == WHIZARD_ROOT) &
            & call read (uraw, mcs%evt, fmt=RAW_FMT, iostat=iostat)
!       call mpi90_broadcast (mcs%evt, WHIZARD_ROOT)
       call mpi90_broadcast (iostat, WHIZARD_ROOT)
       if (iostat == 0) then
          mcs%evt%count = mcs%evt%count + 1
       else if (iostat < 0) then
          call reopen_raw_events_file
       else
          stop "I/O error occured while reading raw event file"
       end if
    end if
    if (.not.mcs%read_events_raw) then
       call mc_generate (excess=excess)
       if (mcs%write_events_raw) then
          if (proc_id == WHIZARD_ROOT) call write (uraw, mcs%evt, fmt=RAW_FMT)
       end if
    end if
    call hepevt_fill (mcs%evt)
    call apply (whizard_analysis, mcs%evt, excess=excess)
    if (mcs%fragment) then
       select case (mcs%fragmentation_method)
       case (JETSET)
          call jetset_fragment
       end select
       if (proc_id == WHIZARD_ROOT .and. mcs%write_events) &
            & call hepevt_write (uevt, fmt=mcs%write_events_format)
    else
       if (proc_id == WHIZARD_ROOT .and. mcs%write_events) &
            & call write (uevt, mcs%evt, fmt=mcs%write_events_format)
    end if
  contains
    subroutine reopen_raw_events_file
      mcs%read_events_raw = .false.
      if (proc_id == WHIZARD_ROOT) then
         close (uraw)
         print *, "! Warning: End of file reached after reading", &
              & mcs%evt%count, "events"       
         print *, "! Generating fresh events ..."
         if (mcs%write_events_raw) then
            open(unit=uraw, file=trim(whizard_filename)//".evx", &
                 & status='old', position='append', &
                 & action='write', form='unformatted')
         end if
      end if
    end subroutine reopen_raw_events_file
  end subroutine whizard_event

  subroutine whizard_events_end
    if (.not.event_files_open)  return
    mcs%it = mcs%it + 1
    call mpi90_rank (proc_id)
    if (proc_id == WHIZARD_ROOT) then
       call write (6, whizard_analysis, mcs, &
            &      show_histograms=mcs%screen_histograms)
       call write (trim(whizard_filename)//"-plots", TEXT_FMT, &
            &      whizard_analysis, mcs)
       call write (trim(whizard_filename)//"-plots", GML_FMT, &
            &      whizard_analysis, mcs)
    end if
    call close_raw_events_file
    call close_output_events_file
    event_files_open = .false.
  contains
    subroutine close_raw_events_file
      if (proc_id == WHIZARD_ROOT) then
         if (mcs%read_events_raw .or. mcs%write_events_raw)  close (uraw)
      end if
    end subroutine close_raw_events_file
    subroutine close_output_events_file
      if (proc_id == WHIZARD_ROOT .and. mcs%write_events) then
         select case (mcs%write_events_format)
         case (JETSET_FMT1, JETSET_FMT2, JETSET_FMT3)
         case (STDHEP_FMT)
            call stdhep_end
         case default
            close (uevt)
         end select
      end if
    end subroutine close_output_events_file
  end subroutine whizard_events_end

  subroutine generate_events_weighted (n_calls)
    integer, intent(in) :: n_calls
    integer :: ievt, ch, iostat
    real(kind=default) :: weight, efftmp
    real(kind=default), dimension(:), allocatable :: eff, effsum
    type(histogram), dimension(:), allocatable :: hh
    type(histogram) :: h
    allocate (hh(mcs%f%n_channels))
    call create_histogram (h, 0._default, 1._default, WEIGHT_HISTOGRAMS_NBIN)
    allocate (eff(mcs%f%n_channels), effsum(mcs%f%n_channels))
    effsum = 0
    uwgt = free_unit ()
    open (unit=uwgt, file=trim(whizard_filename)//".wgt", &
         & status='replace', action='write')
    write (uwgt,*) "! WHIZARD run for process ", trim(mcs%process_id)
    write (uwgt,*) "! Weight distribution"
    do ch=1, mcs%f%n_channels
       if (mcs%g%num_calls(ch) < 2) cycle
       write(uwgt,*) "!"
       write(uwgt,*) "! Channel #", ch
       call create_histogram &
            & (hh(ch), 0._default, 1._default, WEIGHT_HISTOGRAMS_NBIN)
       do ievt=1, int(n_calls * mcs%g%weights(ch))
          if (mcs%read_events_raw) then
             mcs%evt%count = ievt
             if (proc_id == WHIZARD_ROOT) &
                  & call read(uraw, mcs%evt, fmt=RAW_FMT, &
                  &           probability=eff(ch), weight=weight, channel=ch, &
                  &           iostat=iostat)
!             call mpi90_broadcast (mcs%evt, WHIZARD_ROOT)
             if (iostat == 0) then
                mcs%evt%count = mcs%evt%count + 1
             else if (iostat < 0) then
                stop "Unexpected end of raw event file"
             else
                stop "I/O error occurred while reading raw event file"
             end if
          end if
          if (.not.mcs%read_events_raw) then
             call mc_generate(probability=eff(ch), weight=weight, channel=ch)
             if (mcs%write_events_raw) then
                 if (proc_id == WHIZARD_ROOT) &
                      & call write(uraw, mcs%evt, fmt=RAW_FMT, &
                      &            probability=eff(ch), weight=weight, &
                      &            channel=ch)
              end if
           end if
           effsum(ch) = effsum(ch) + eff(ch)
           call fill_histogram (hh(ch), eff(ch))
           call fill_histogram (h, eff(ch))
           call apply(whizard_analysis, mcs%evt, weight/mcs%integral)
        end do
        if (proc_id == WHIZARD_ROOT) then
           call write_histogram (hh(ch), uwgt, over=.true.)
           write(uwgt,*) "!"
           if (int(n_calls * mcs%g%weights(ch)) > 0) then
              efftmp = effsum(ch) / int(n_calls * mcs%g%weights(ch))
           else
              efftmp = 0
           end if
           write(uwgt,*) "! Efficiency = ", efftmp
        end if
     end do
     if (proc_id == WHIZARD_ROOT) then
        write(uwgt,*) "!"
        write(uwgt,*) "! Multi-channel:"
        call write_histogram (h, uwgt, over=.true.)
        write(uwgt,*) "!"
        write(uwgt,*) "! Warmup events    = ", mcs%calls(1,4)*mcs%calls(2,4)
        write(uwgt,*) "! Generated events = ", n_calls
        write(uwgt,*) "! Efficiency = ", sum(effsum) / n_calls
     end if
     close (uwgt)
     deallocate (hh)
     deallocate (eff, effsum)
  contains
    subroutine reopen_raw_events_file
      mcs%read_events_raw = .false.
      if (proc_id == WHIZARD_ROOT) then
         close (uraw)
         print *, "! Warning: End of file reached after reading", &
              & mcs%evt%count, "events"       
         print *, "! Generating fresh events ..."
         if (mcs%write_events_raw) then
            open(unit=uraw, file=trim(whizard_filename)//".evx", &
                 & status='old', position='append', &
                 & action='write', form='unformatted')
         end if
      end if
    end subroutine reopen_raw_events_file
  end subroutine generate_events_weighted

  subroutine whizard_end
    if (event_files_open)  call whizard_events_end
    if (mcs_exists) then
       call destroy (mcs)
       mcs_exists = .false.
    end if
    if (whizard_input_exists) then
       call destroy (whizard_input)
       whizard_input_exists = .false.
    end if
    print *, "! WHIZARD run finished"
    if (mpi_initialized) then
       call mpi90_finalize
       mpi_initialized = .false.
    end if
  end subroutine whizard_end

end module whizard
